### Script Generator for IP simulators ###
#
# Coded by Yasusi Kanada
# 2015-5-17, 6-27; 2016-6-25

import sys, re, optparse

## Internal data structures ##

class NeighborDict(dict):	# hash table (dictionary) of neighbors
    def __missing__(self, key):
        self[key] = None
        return self[key]

class DeviceDict(dict):		# hash table (dictionary) of devices
    def __missing__(self, key):
        self[key] = { "type": None, "neighbors": NeighborDict(), "numPort": 0 }
        return self[key]

class GatewayDict(dict):       	# hash table (dictionary) of gateways
    def __missing__(self, key):
        self[key] = None
        return self[key]

devices = DeviceDict()

gateway = GatewayDict()


## Options ##
parser = optparse.OptionParser()
parser.add_option("--mac", dest="isMac", action = "store_true")
parser.add_option("--path", dest="path", type = "string")
opts, args = parser.parse_args()


## Binary path ##

path = opts.path
if opts.isMac:
    pythonCommand = "python3"
    if path == None:
        path = "/Library/Frameworks/Python.framework/Versions/3.3/bin"
else:
    pythonCommand = "python"
    if path == None:
        path = "c:\python36;c:\python35;c:\python34;c:\python33;c:\python32;c:\python31"


def getIP(arg):
    IP = arg.split('.')
    if len(IP) != 4:
        print("Error: IP address {} format must be ddd.ddd.ddd.ddd".format(IP))
    return (int(IP[0]) << 24) + (int(IP[1]) << 16) + (int(IP[2]) << 8) + int(IP[3])


def putIP(IP):
    return str(IP >> 24) + '.' + str((IP >> 16) & 255) + '.' + \
        str((IP >> 8) & 255) + '.' + str(IP & 255)

    
## Compile network spec into internal form ##

def genInternal():
    pattern = re.compile(r"\s*(\w+)\s*([\w/\.]+)")

    myDeviceNum = 0
    for line in sys.stdin:    
        m = pattern.match(line)
        if m != None:
            command = m.group(1);
            argument = m.group(2);

            if command == "router" or command == "terminal" or command == "server":
                myDeviceNum = int(argument)
                if myDeviceNum < 100 or 600 < myDeviceNum:
                    print("Error: Device number", myDeviceNum, "must be in 100..600")
                devices[myDeviceNum]["type"] = command
                if command == "router":
                    devices[myDeviceNum]["routing"] = []
                    devices[myDeviceNum]["address"] = GatewayDict()

            elif command == "subnet":
                args = argument.split('/')
                if len(args) < 2:
                    print("Error: subnet argument {} format must be IPaddress/len".\
                          format(argument))
                    print("  (subnet: 0 < len < 32")
                portIP = getIP(args[0])
                netLen = int(args[1])
                myDevice = devices[myDeviceNum]
                myPort = myDevice["numPort"]
                myDevice["numPort"] += 1
                mask = 2 ** 32 + (-1 << (32 - netLen))
                subnet = portIP & mask
                gateway[portIP] = { "net": subnet, "netLen": netLen,
                                    "deviceNum": myDeviceNum, "port": myPort }
                myDevice["address"][subnet] = portIP

            elif command == "gateway":
                gatewayIP = getIP(argument)
                myDevice = devices[myDeviceNum]
                if gateway.get(gatewayIP) == None:
                    print("Error: next hop '{}' not defined".format(putIP(gatewayIP)))
                myGateway = gateway[gatewayIP]
                myDevice["netLen"] = myGateway["netLen"]
                nextDeviceNum = gateway[gatewayIP]["deviceNum"]
                if nextDeviceNum == None:
                    print("Error: next hop '{}' not defined".\
                          format(putIP(gateway[gatewayIP])))

                myPort = myDevice["numPort"]
                myDevice["numPort"] += 1
                myNeighbors = myDevice["neighbors"]

                nextDevice = devices[nextDeviceNum]
                nextPort = nextDevice["numPort"]
                nextDevice["numPort"] += 1
                nextNeighbors = nextDevice["neighbors"]

                myNeighbors[nextDeviceNum] = { "myPort": myPort, "nextPort": nextPort }
                nextNeighbors[myDeviceNum] = { "myPort": nextPort, "nextPort": myPort }

            elif command == "localIP" or command == "remoteIP":
                IPaddr = getIP(argument)
                devices[myDeviceNum][command] = IPaddr

            elif command == "static":
                route = argument.split("/")
                if len(route) < 3:
                    print("Error: route argument '{}' format must be subnet/len/next_hop_IP".\
                          format(argument))
                    print("  (0 < len < 32, next_hop_IP: 32-bit")
                subnet = getIP(route[0])
                netLen = int(route[1])
                nexthopIP = getIP(route[2])
                devices[myDeviceNum]["routing"].\
                    append({ "subnet": subnet, "netLen": netLen, "nexthopIP": nexthopIP })


## Generate scripts from internal form ##

def genScripts():
    for myDeviceNum in devices.keys():
        myDevice = devices[myDeviceNum]
        f = open("{}{}.bat".format(myDevice["type"], myDeviceNum),
                 "w", encoding="utf-8")

        if opts.isMac:
            print("PATH={}".format(path), file=f)
        else:
            print("path", path, file=f)

        neighbors = myDevice["neighbors"]
        if myDevice["type"] == "router":
            routeCount = 0
            print(pythonCommand, "srouter.py --monitor 0 --dumpIP", end="", file=f)
            for nextDeviceNum in myDevice["neighbors"].keys():
                next = neighbors[nextDeviceNum]
                nextDevice = devices[nextDeviceNum]
                print(" --route{} {}/{}/{}/{}/{}".\
                      format(routeCount, nextDevice["localIP"], nextDevice["netLen"],
                             100 * myDeviceNum + next["myPort"],
                             100 * nextDeviceNum + next["nextPort"], "*"),
                      end="", file=f)
                routeCount += 1
            for route in myDevice["routing"]:
                nexthopIP = route["nexthopIP"]
                nexthopRouter = gateway[nexthopIP]
                if nexthopRouter == None:
                    print("Error: unknown next hop address {}".format(putIP(nexthopIP)))
                    exit()
                nexthopMask = 2 ** 32 + (-1 << (32 - nexthopRouter["netLen"]))
                nexthopNet = nexthopIP & nexthopMask
                portIP = myDevice["address"][nexthopNet]
                if portIP == None:
                    print("Error: unknown next hop address {}".format(putIP(nexthopIP)))
                    exit()
                thisRouter = gateway[portIP]
                if thisRouter == None:
                    print("Error: unknown local address {}".format(putIP(portIP)))
                    exit()
                thisPort = 100 * thisRouter["deviceNum"] + thisRouter["port"]
                nexthopPort = 100 * nexthopRouter["deviceNum"] + nexthopRouter["port"]
                print(" --route{} {}/{}/{}/{}/{}".\
                      format(routeCount, route["subnet"], route["netLen"],
                             nexthopPort, thisPort, nexthopIP),
                      end="", file=f)
                routeCount += 1
            print(file=f)

        elif myDevice["type"] == "terminal":
            if myDevice.get("localIP") == None:
                print("Error: Local IP address not specified for terminal",
                      myDeviceNum)
            if myDevice.get("remoteIP") == None:
                print("Error: Remote IP address not specified for terminal",
                      myDeviceNum)
            if myDevice["numPort"] > 1:
                print("Error: Two or more links specified for terminal", myDeviceNum)
            for nextDeviceNum in myDevice["neighbors"].keys():
                next = neighbors[nextDeviceNum]
                print(pythonCommand, "iterm.py --lip", myDevice["localIP"],
                      "--rip", myDevice["remoteIP"],
                      "--lp", 100 * myDeviceNum + next["myPort"],
                      "--rp", 100 * nextDeviceNum + next["nextPort"], file=f)

        elif myDevice.get("type") == "server":
            if myDevice.get("localIP") == None:
                print("Error: Local IP address not specified for terminal",
                      myDeviceNum)
            if myDevice["numPort"] > 1:
                print("Error: Two or more links specified for terminal", myDeviceNum)
            for nextDeviceNum in myDevice["neighbors"].keys():
                next = neighbors[nextDeviceNum]
                print(pythonCommand, "iserver.py --lip", myDevice["localIP"],
                      "--lp", 100 * myDeviceNum + next["myPort"],
                      "--rp", 100 * nextDeviceNum + next["nextPort"], file=f)
        else:
            print("Error: Unknown device type:", myDevice["type"])

        print("pause >nul", file=f)
        f.close()


genInternal()

genScripts()
