### IP router (static routing only) simulator (UDP w/o select) ###
#
# Public domain doftware
# (When using Python 2, part of this program must be commented/uncommented)
#
# Coded by Yasusi Kanada
# Initial version: 2015-5-17
# Updated: 2017-6-10  (not counting packets within a subnet)

import optparse, socket, time
from ip import SIP, DIP, Payload, printPacket, hex2decIPv4

BUF_SIZE = 3100		# > 2 * 1500

host = socket.gethostbyname(socket.gethostname())


def formatIP(addr):
    return str(addr >> 24) + "." + str((addr >> 16) & 255) + "." + \
        str((addr >> 8) & 255) + "." + str(addr & 255)


class PacketCountDict(dict):	# hash table (dictionary) for packet counts
    def __missing__(self, key):
        self[key] = 0
        return self[key]

packetCount = PacketCountDict()


### Packet handler ###

vport2remote = {}

def sendPacket(packet, localPort):
    sock, addr = vport2remote[localPort]
    return sock.sendto(packet, addr) // 2

def printForwarding(message, dIP):
    print("{} {} {} {} {}".format("@@@", message, "- Dest. IP:", dIP, "@@@"))

def printSending(sendBytes):
    print("{} bytes sent".format(sendBytes))


### Routing table handler ###

routelist = []
routedict = {}

def setRoute(subnet, netLen, localPort, remotePort, nexthop):
    mask = 2 ** 32 + (-1 << (32 - int(netLen)))
    subnet = subnet & mask
    if routedict.get(subnet) == None:
        route = { "subnet": subnet, "mask": mask, "links": [], "nexthop": nexthop }
        routelist.append(route)
        routedict[subnet] = route
    else:
        route = routedict[subnet]
    route["links"].append({ "localPort": localPort, "remotePort": remotePort })

def searchRoute(address):
    for route in routelist:
        net = address & route["mask"]
        if net == route["subnet"]:	# matched!
            return route
    return None		# 170610

def dumpRoutes():
    print("\n\n\n\n\n\n\n\n\n\n\n\n")
    print("{:16} {:16} {:16} {:12}".\
          format("Subnet", "Mask", "NextHop", "Packet count"))
    for route in routelist:
        nexthop = route["nexthop"]
        if nexthop != "*":
            nexthop = formatIP(nexthop)
        print("{:16} {:16} {:16} {:12}".\
              format(formatIP(route["subnet"]), formatIP(route["mask"]), \
                     nexthop, packetCount[route["subnet"]]))
    if len(routelist) == 0:
        print("Empty")


### Packet forwarding ###

def forward(packet):
    now = time.time()
    if opts.monitorLevel >= 1:
        printPacket(packet, now, "Received:")

    # Forward packet
    addr = int(DIP(packet), 16)
    route = searchRoute(addr)
    for link in route["links"]:
        sendBytes = sendPacket(packet, link["localPort"])

    saddr = int(SIP(packet), 16)	# 170610
    sroute = searchRoute(saddr)		# 170610
    if route["subnet"] != sroute["subnet"]:	# 170610
        packetCount[route["subnet"]] += 1
        
    if opts.monitorLevel >= 2:
        printSending(sendBytes)


### Main program ###

def main():
    global opts

    # Parse command line
    parser = optparse.OptionParser()
    for i in range(0, 8):
        str_i = str(i);
        parser.add_option("--route" + str_i, dest = "route" + str_i,
                          type = "str", help = ("Route #" + str_i))
    parser.add_option("--monitorLevel", "--monitor",
                      dest = "monitorLevel", type="int",
                      help = ("Monotoring level (0: Silent, 1: Sending/Receiving 2: Debug)"))
    parser.add_option("--dumpIP", dest = "dumpIP", action = "store_true")
    opts, args = parser.parse_args()

    # Create route table
    routes  = [opts.route0, opts.route1, opts.route2, opts.route3,
               opts.route4, opts.route5, opts.route6, opts.route7]

    # Set socket, port mapper, and routing table
    sock = []
    vport = []
    nports = 0
    for i in range(0, len(routes)):
        if routes[i] != None:
            rs = routes[i].split("/")
            if len(rs) >= 4:
                (ssubnet, snetLen, slocalPort,
                 sremotePort, nexthop) = rs
                localPort = int(slocalPort)
                remotePort = int(sremotePort)
                if nexthop != "*":
                    nexthop = int(nexthop)

                if vport2remote.get(localPort) == None:
                    setRoute(int(ssubnet), int(snetLen),
                             localPort, remotePort, nexthop)

                    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    s.setblocking(0)
                    s.bind((host, localPort))
                    sock.append(s)

                    vport.append(localPort)
                    vport2remote[localPort] = (sock[nports], (host, remotePort))

                    nports += 1
            else:
                print("System error: too few argument elements {}".format(rs))

    # Repeat forwarding a packet
    interval = 1
    lastTime = time.time()
    while True:
        for i in range(0, nports):
            sourceSock = sock[i]
            try:
                packet = sourceSock.recv(BUF_SIZE)
            except socket.error:
                pass
            else:
                forward(packet)

        now = time.time()
        if now > lastTime + interval:
            if opts.dumpIP:
                dumpRoutes()
            lastTime = now
        else:
            time.sleep(0.001001)

main()
